package org.kne.cloud.network.perf;

import org.kne.cloud.network.monitor.SpeedAndTrafficAndDelayMonitorDataImpl;

public interface KperfReports {
	public void testProcess(double d);
	public void testFinish();
	public void updateDial(SpeedAndTrafficAndDelayMonitorDataImpl mdt);
	public void uploadSpeedFinish(long uploadSpeed);
	public void downloadSpeedFinish(long downloadSpeed);
	public void uploadPPSFinish(long uploadPPS);
	public void downloadPPSFinish(long downloadPPS);
	public void uploadDelayFinish(long uploadDelay);
	public void downloadDelayFinish(long downloadDelay);
}
